/****************************************************************************** 
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
 // Summary:
 //    The SectionLayoutForm displays an object's (content) attributes in sections. Each section
 //    collects a set of related attributes for the content. The form displays sections in tiled layout,
 //    controlled by settable attributes such as the layout direction and the number of vertical columns in the form. 
 //    Each sections has a list of attributes. For "simple", or atomic attributes, the attribute has a label
 //    and if the form is editable, an widget to display and edit the value of the attribute. The presentation widget
 //    for the attribute is determined by the attribute's value types, such as text, date, enumerated values, etc.
 //
 //    The SectionLayoutForm does not directly know the structure of the content that is being displayed. Instead,
 //    it relies on an intermediator object called an provider to query the layout information of the content. The 
 //    provider supplies the content information to the form layout. This decouples the core form layout function
 //    from the contents that it displays, and this makes the form layout more generic and allows it to display other
 //    contents without any internal changes. To display another content type, the application simply create a
 //    provider object that intimately knows the content type and set it to the form layout. Here's a general usage
 //    model pseudo code:
 //
 //       // Create a concrete implementation provider class.  
 //       this.provider = new AssetDetailView.TestProvider();
 //       this.formLayout = new com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm(this.viewHeaderNode.getContentNode());
 //       this.formLayout.setContentProvider(this.provider);
 //       this.formLayout.setContent(dummyContent);
 //       this.formLayout.setListener(new FormListener());
 //
 //       // create the form's content 
 //       this.formLayout.doLayout();
 //
 //       // load the content's attribute values to the form.
 //       this.formLayout.loadDataToForm();
 //    
 //    Please note that SectionLayoutForm is not a widget but rather a simple object type. It needs a root dom object
 //    to build the form from.
 //
SectionLayoutForm = dojo.provide("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm");

SectionLayoutForm.TILE_VERTICAL             = 1;
SectionLayoutForm.TILE_HORIZONTAL           = 2;

SectionLayoutForm.ATTRIBUTE_LABEL           = 3;
SectionLayoutForm.ATTRIBUTE_TEXT            = 4;
SectionLayoutForm.ATTRIBUTE_TEXT_NUMERIC    = 5;
SectionLayoutForm.ATTRIBUTE_TEXT_DATE       = 6;
SectionLayoutForm.ATTRIBUTE_TEXT_IP         = 7;
SectionLayoutForm.ATTRIBUTE_ENUM            = 8;
SectionLayoutForm.ATTRIBUTE_COLLECTION      = 9;
SectionLayoutForm.ATTRIBUTE_CUSTOM          = 10;
SectionLayoutForm.ATTRIBUTE_TEXT_CHECKBOX   = 11;

dojo.require("dojox.data.dom");
dojo.require("net.jazz.ajax.toolbox.lang.internal.Lang");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.RtlmUiUtil");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionHeader");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormDateField");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormLabelField");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormTextField");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormTextCheckboxField");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormEnumField");
dojo.require("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormListener");

(function() {
    
// Convenient local shorthand variables
//
var SectionHeader = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionHeader;
var checkNotNull    = net.jazz.ajax.toolbox.lang.checkNotNull;
var checkNotEmpty   = net.jazz.ajax.toolbox.lang.checkNotEmpty;

var LabelField      = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormLabelField;
var TextField       = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormTextField;
var TextCheckboxField = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormTextCheckboxField;
var EnumField       = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormEnumField;
var DateField       = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormDateField;
var NumericField    = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormNumericField;
var IPAddressField  = com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormIPAddressField;


var MapStructId = "id";
var MapStructInfo = "info";
var MapStructObjectSecondKey = "secondKey";
var MapStructObjectKey = "objectKey";
var MapStructRootNode = "rootNode";
var MapStructField = "node";


// Summary:
//    Defines an interface for which the SectionLayoutForm uses to query the layout
//    information for the content object. The SectionLayoutForm has no direct knowledge about
//    the object of which the form is being constructed. The application (consumer) of the 
//    SectionLayoutForm needs to supply a provider that is specific to the contents being 
//    layout. The supplied provider must be a subclass of this base class and overrides
//    all methods defined in this base class.
//
dojo.declare("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm.ContentProvider", null,
{    
    // Summary:
    //    Returns the layout information about the form. The following is the info format:
    //    {
    //         nColumns: integer > 0,  the number of sections across the form layout. 
    //         tileDirection: value,   SectionLayoutForm.TILE_VERTICAL, or SectionLayoutForm.TILE_HORIZONTAL
    //         sections: ["key",...]   the array of section keys** (please see below)
    //    }
    //    If the tiling direction is vertical, the sections in the info object are filled from
    //    column 1 first. Otherwise, for horizontal tiling, the sections are filled from
    //    row 1 first.
    //
    //    ** The section keys can be a simple string tag, or an opique object. If it's an opique
    //    object, it must have an id attribute, like key.id that is a string value. The key (or id) must be
    //    unique amoung section ids. The section's key (id) can then be used to reference and access the
    //    section layout.
    //
    getFormInfo: function(/* the content object */content) 
    {    
    },
    
    // Summary:
    //    Returns information about a section given its key value. The following is the info format:
    //    {
    //         title: "string",          the title for the section
    //         nColumns: integer > 0,    the number of columns of attributes across the section. 
    //         attributes: ["key",...]   the array of attribute keys** (please see below)
    //    }
    //
    //    ** The attribute keys can be a simple string tag, or an opique object. If it's an opique
    //    object, it must have an id attribute, like key.id that is a string value. The key (or id) must be
    //    unique among attribute ids. The attribute's key (id) can then be used to reference and access the
    //    attribute layout.
    //
    getSectionInfo: function(/* the content object */content, sectionKey)
    {
    },
    
    // Summary:
    //    Returns information about the attribute given its key value. The attribute info is not static for
    //    all attributes being described. Certain attribute types will have additional data 
    //    associated with them.
    //
    //    The following is the info format:
    //    {
    //         /* 
    //          * Required Info Data 
    //          */
    //         type: value,               the type of attribute, one of the SectionLayoutForm.ATTRIBUTE_*
    //
    //         /* 
    //          * Optional Info Data
    //          */
    //         colSpan: <1..section.nCol> Optionally overrides the parent section layout's nCols attribute.
    //                                    By default, the attribute will occupy one column of the section. This
    //                                    allows the attribute to occupy more then one column in the section.
    //                                    For attribute type ATTRIBUTE_COLLECTION, the default is section.nCols,
    //                                    if no colSpan is specified. It occupies the entire row in the section.
    //         canDelete: <true,false>,   Optional, if set to true, a "delete" button is placed at the right of
    //                                    field. The delete button will generate an event to the form listener.
    //                                    It's up the listener/application to remove the field from the form.
    //
    //         /* 
    //          * Common Info Data for "atomic" attribute types
    //          */
    //         title: "string",          the title for the attribute
    //         readOnly: <true,false>,  
    //         isRequired: <true,false>,
    //         tooltipMessage: "string", tooltip for the field, currently only for FormTestCheckboxField
    //
    //         /* 
    //          * Specific Info Data for the SectionLayoutForm.ATTRIBUTE_COLLECTION type
    //          */
    //         collection: [keys...],    the array of sub-attribute keys** (please see below)
    //         title: "string",          Optional, the title for the collection, if null no title (row) will be created
    //         indentLevel: 0..5,        Optional, provide an absolute indentation level to the right of the form, if none is specified, no indentation
    //         nColumns: integer > 0     Optional, the number of attributes in a row. This overrides the parent section's nColumns value.
    //
    //         /* 
    //          * Specific Info Data for the SectionLayoutForm.ATTRIBUTE_ENUM type
    //          */
    //         enumChoices:[values...]
    //         createFirstEmpty: <true,false> if set to true, a first empty choice is always created
    //    }
    //
    //    ** The attribute keys can be a simple string tag, or an opique object. If it's an opique
    //    object, it must have an id attribute, like key.id that is a string value. The key (or id) must be
    //    unique among ALL attribute ids. The attribute's id can then be used to reference and access the
    //    attribute layout.
    //
    getAttributeInfo: function(/* the content object */content, sectionKey, attributeKey)
    {
    },
    
    // Summary:
    //    Returns the value of the attribute given the attrubute key. 
    //
    getAttributeValue: function(/* the content object */content, attributeKey)
    {
    }
});

// Summary:
//    Defines an interface for which the SectionLayoutForm uses to obtain custom elements 
//    to be created for the form. Custom elements are items that are specific to the application
//    and thus must be created by the application (consumer of the form). If a CustomElementProvider
//    is set to the SectionLayoutForm object, the form layout will query it for creating the parts
//    in the form. Specifically, two types of custom elements can by supplied to the SectionLayoutForm:
//    1) section header, 2) attribute fields.
//
//    NOTE: In the current form, constructSectionHeader must return a section header that is sub-classed from
//          com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionHeader.  If this is not the case
//          then the application will "crash" when it goes to set the content on the header, if the contentNode
//          does not exist.  If this design does not have enough flexibility, please bring this forward.
//
dojo.declare("com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm.CustomElementProvider", null,
{
    // Summary:
    //    Returns a widget that is the section header.
    //
    constructSectionHeader: function(sectionKey)
    {
    },
    getSectionHeaderContentNodeName: function()
    {     
    },
    // Summary:
    //    Returns a widget that is the header for collection attributes.
    //
    constructCollectionAttributeHeader: function(sectionKey, atttributeKey)
    {
    },
    // Summary:
    //    Returns a widget for the attribute. The widget must be a subclass of the AbstractFormField.
    // 
    constructCustomField: function(atttributeKey)
    {
    }
});
   


dojo.declare(
  "com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm", null,
{
        
    // Summary:
    //    The provider object of the "interface" type: com.ibm.rational.test.lm.web.ui.internal.widgets.SectionLayoutForm.ContentProvider 
    //    Consumer must supply a provider object that implements the methods defined in the object's class.
    //
    provider: null,
    
    // Summary:
    //    The provider object of the "interface" type: com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm.CustomElementProvider
    //    Consumer of the form can optionally provide a custom element provider to override the contents being created in the form.
    //
    elementProvider: null,
    
    // Summary:
    //    The object of which the section form layout is created for and edited by.
    //
    content: null,
    
    // Summary:
    //    The listener object for form events.
    //
    formListener: null,
    
    // Summary:
    //    The root dom node of the form.
    //
    rootNode: null,
    
    // Summary:
    //    For every section/attribute created in the form, an entry holding the layout data
    //    for the item is stored in the respective maps, "hashed" by its id.
    //
    sectionMap: null,
    attributeMap: null,
    
    // Summary:
    //    Constructs a SectionLayoutForm. A root node of which the form to be build from is required.
    //
    constructor: function(rootNode)
    {
        this.rootNode = rootNode;
    },
    
    // Summary:
    //    Registers a listener to receive form events. Appropriate messages (the listener's methods)
    //    will be fired when certain form events occurred. The listener must be of the type:
    //    com.ibm.rational.test.lm.web.ui.internal.widgets.layout.FormListener.
    //
    setListener: function(listener)
    {
        this.formListener = listener;
    },
    
    // Summary:
    //    Loads the form's structural content. 
    //
    setContent: function(content)
    {
        this.content = content;
    },
    
    // Summary:
    //    Returns the form's content object. 
    //
    getContent: function()
    {
        return this.content;
    },
    
    // Summary:
    //
    setContentProvider: function(provider)
    {
        this.provider = provider;
    },

    // Summary:
    //
    getContentProvider: function()
    {
        return this.provider;
    },

    // Summary:
    //
    setCustomElementProvider: function(provider)
    {
        this.elementProvider = provider;
    },

    // Summary:
    //
    getCustomElementProvider: function()
    {
        return this.elementProvider;
    },
    
    // Summary:
    //    Returns whether the form contains empty required field values. Returns true
    //    if any field is a required field and it has an empty value in the field.
    //
    hasEmptyRequiredFields: function()
    {
        var hasEmptyReq = false;
        
        for(var attrKey in this.attributeMap) {
            var attrRec = this.attributeMap[attrKey];
            var field = attrRec[MapStructField];
            
            if(field && field.getIsRequired()) {
                var val = field.getValue();
                if(!val || (dojo.isString(val) && val.length == 0)) {
                    hasEmptyReq = true;
                    break;
                }
            }
        }
        return hasEmptyReq;        
    },
    
    // Summary:
    //    Returns whether the form has fields that has erroneous value entered. Returns
    //    true if any field has error value.
    //
    hasErrorFields: function()
    {
        var hasError = false;
        
        for(var attrKey in this.attributeMap) {
            var attrRec = this.attributeMap[attrKey];
            var field = attrRec[MapStructField];
            
            if(field && field.getHasError()) {
                hasError = true;
                break;
            }
        }
        return hasError;     
    },

    // Summary:
    //    Returns whether there're fields that has been edited. If one field is edited, this method returns true.
    //
    hasEdited: function()
    {
        var hasEditing = false;
        
        for(var attrKey in this.attributeMap) {
            var attrRec = this.attributeMap[attrKey];
            var field = attrRec[MapStructField];
            
            if(field && field.getIsEdited()) {
                hasEditing = true;
                break;
            }
        }
        return hasEditing;
    },
    
    // Summary:
    //    Clear the fields' edited state. This makes the form field to have the non-edited appearance.
    //
    clearAllEditedState: function()
    {
        for(var attrKey in this.attributeMap) {
            var attrRec = this.attributeMap[attrKey];
            var field = attrRec[MapStructField];
            
            if(field && field.getIsEdited()) {
                field.setEdited(false);
            }
        }
        
    },
    
    // Summary:
    //    Constructs the form based on the supplied content and providers. If any of the required
    //    member are missing, i.e., the content, the provider, or the root node ref, this method
    //    do nothing.
    
    doLayout: function()
    {
        if(!this.content || !this.provider || !this.rootNode)
            return;
            
        //if(!(this.provider instanceof com.ibm.rational.test.lm.web.ui.internal.widgets.layout.SectionLayoutForm.ContentProvider)) {
        //    return;
        //}
        
        // initialize the maps for the layout items.
        this.sectionMap = {};
        this.attributeMap = {};
        
        dojox.data.dom.removeChildren(this.rootNode);
        var mainDiv = document.createElement("div");
        mainDiv.className = "section-layout-form";
        mainDiv.tabIndex = -1;
        
        this.rootNode.appendChild(mainDiv);
        this.rootTable = document.createElement("table");
        
        this.rootTable.className = "root-table";
        this.rootTable.tabIndex = -1;
        mainDiv.appendChild(this.rootTable);
        
        this._createSectionLayout(this.rootTable);
    },
    
    // Summary:
    //    Loads values into the form's attribute fields. The form queries the ContentProvider 
    //    for each of attribute's value in the form. If there's value in the fields, those values
    //    will be replaced. No event is generated by this method.
    // 
    loadDataToForm: function()
    {
        this.inLoading = true;
        try {
            for(var attrKey in this.attributeMap) {
                var attrRec = this.attributeMap[attrKey];
                var field = attrRec[MapStructField];
                
                if(field) {
                    var value = this.provider.getAttributeValue(this.content, attrRec[MapStructObjectKey]);
                    if(value)
                        field.setValue(value);
                }
            }
        }
        catch(err) {
            console.debug(err);
        }
        this.inLoading = false;
    },
    
    recurseLoadAttributes: function(attrKeys)
    {
        this.inLoading = true;
        try {
            this._recurseLoadValue(attrKeys);
        }
        catch(err) {
            console.debug(err);
        }
        this.inLoading = false;
    },
    
    // Summary:
    //    Given a section identified by its key, returns the list of its children attribute keys.
    //    These are the collection of attributes that is returned by the form's content provider
    //    via the sectionInfo data record. The form retains this information.
    //
    getSectionAttributes: function(sectionKey)
    {
        var id = this._getKeyId(sectionKey);
        var sectionData = this.sectionMap[id];
        
        if (sectionData) {
            return sectionData[MapStructInfo].attributes;
        }
        return null;
    },
    
    // Summary:
    //    Deletes all elements in the section identified by the given sectionKey. Recreates the section's
    //    contents by querying the contentProvider again.
    //
    invalidateSectionLayout: function(sectionKey)
    {
        var id = this._getKeyId(sectionKey);
        var section = this.sectionMap[id];
        
        if(section) {
            dojox.data.dom.removeChildren(section[MapStructRootNode]);
            this._createSectionContent(section[MapStructRootNode], section[MapStructObjectKey]);
            
            // reload the value in this section
            var sectionInfo = this.provider.getSectionInfo(this.content, section[MapStructObjectKey]);
            if(sectionInfo) {
                this.recurseLoadAttributes(sectionInfo.attributes);                
            }
        }
        else {
            console.debug("invalidateSectionLayout: Section not registered in the form: " + id);
        }
    },
    
    // Summary:
    //    Deletes the field elements identified by the given sectionKey. Recreates the attribute's
    //    contents by querying the contentProvider again.
    //
    invalidateAttributeLayout: function(attributeKey)
    {
        var id = this._getKeyId(attributeKey);
        var attrRec = this.attributeMap[id];
        
        if(attrRec) {
            dojox.data.dom.removeChildren(attrRec[MapStructRootNode]);
            var attrNode = this._createAttribute(attrRec[MapStructRootNode], attrRec[MapStructObjectSecondKey], attrRec[MapStructObjectKey]);
            attrRec[MapStructRootNode].appendChild(attrNode);
            
            // reload the value in this attribute
            var key = [];
            key.push(attributeKey);
            this.recurseLoadAttributes(key);
        }        
        else {
            console.debug("invalidateAttributeLayout: Attribute not registered in the form: " + id);
        }
    },
    
    // Summary:
    //    Returns the form's field value indentified by the attributeKey
    // 
    getAttributeValue: function(attributeKey)
    {
        var id = this._getKeyId(attributeKey);
        var attrRec = this.attributeMap[id];
        
        if (attrRec) {
            var field = attrRec[MapStructField];
            if(field) {
                return field.getValue();
            }
        }
        else {
            console.debug("getAttributeValue: Attribute not registered in the form: " + id);
        }
    },
    
    _recurseLoadValue: function(attrKeys)
    {
        for(var i = 0; i < attrKeys.length; i++ ) {
            var id = this._getKeyId(attrKeys[i]);
            var attrRec = this.attributeMap[id];
            
            if(attrRec) {
                var field = attrRec[MapStructField];
            
                if(field) {
                    var value = this.provider.getAttributeValue(this.content, attrKeys[i]);
                    if(value)
                        field.setValue(value);
                }
                var attrInfo = this.provider.getAttributeInfo(this.content, attrRec[MapStructObjectSecondKey], attrKeys[i]);
                if(attrInfo && attrInfo.type == SectionLayoutForm.ATTRIBUTE_COLLECTION) {
                    this._recurseLoadValue(attrInfo.collection);
                }
            }
        }
    },
    
    _createSectionLayout: function(table)
    {
        var formInfo = this.provider.getFormInfo(this.content);
        var nCols = formInfo.nColumns;
        var sections = formInfo.sections;
        var nSections = sections.length;
        var tileDir = formInfo.tileDirection;
        
        if(nSections == 0)
           return;
        
        if(nCols == 0) {
            console.debug("SectionLayoutForm: cannot have 0 column in form, default to 1 column.")
            nCols = 1;
        }
        
        var reminder = ((nSections % nCols) > 0) ? 1 : 0;
        var nRows = Math.floor(nSections / nCols) + reminder;
        var sect_col_width = Math.floor(100 / nCols) + "%";
        
        for(var i = 0; i < nRows; i++) {
            var tr = table.insertRow(i);
            tr.tabIndex = -1;

            for(var j = 0; j < nCols; j++) {
                var sectionNum = (tileDir == SectionLayoutForm.TILE_HORIZONTAL) ? ((i * nCols) + j) : (i + (j * nRows));
                
                // no more sections to layout.
                if(sectionNum >= nSections) {
                    break;
                }
                
                var sectionKey = sections[sectionNum];
                var sectId = this._getKeyId(sectionKey);
                var td = tr.insertCell(j);
                
                td.tabIndex = -1;
                td.id = "section_" + sectId;
                td.width = sect_col_width;
                this._createSectionContent(td, sectionKey);
            }
        }
    },
    
    _createSectionContent: function(td, sectionKey)
    {
        var sectionInfo = this.provider.getSectionInfo(this.content, sectionKey);        
        var sectHeader;
        var contenNodeName = 'contentNode';
        
        td.tabIndex = -1;
        var contentNodeName;            
        if(this.elementProvider == null)
            sectHeader = new SectionHeader({title: sectionInfo.title});
        else {
            sectHeader = this.elementProvider.constructSectionHeader(sectionKey);
            contentNodeName = this.elementProvider.getSectionHeaderContentNodeName();
        }
                
        var attrKeys = sectionInfo.attributes;
        var nAttrs = attrKeys.length;
        var nCols = sectionInfo.nColumns;
        
        // create the section map for the section.
        var sectionData = this._createLayoutData(td, sectionKey, sectionInfo);
        this.sectionMap[sectionData.id] = sectionData;

        if(nCols == 0) {
            console.debug("SectionLayoutForm: cannot have 0 column in section, default to 1 column.")
            nCols = 1;
        }

        var table = this._createAttributesLayout(sectionKey, attrKeys, nAttrs, nCols);
        
        // Add content to the section header so that it can be collapsed easily without
        // knowing needing to know anything except its internals.
        if(contentNodeName && sectHeader[contentNodeName])
            sectHeader[contentNodeName].appendChild(table);
        else 
            sectHeader.appendChild(table);    
		
		// fix defect 8464	
		sectHeader.toggle();
		sectHeader.toggle();
        
        var node = (sectHeader.domNode) ? sectHeader.domNode : sectHeader;
        td.appendChild(node);
    },
    
    _createAttributesLayout: function(sectionKey, attrKeys, nAttrs, nCols)
    {
        var sectionInfo = this.provider.getSectionInfo(this.content, sectionKey);
        var table = document.createElement("table");
        var sectId = this._getKeyId(sectionKey);
        
        table.className = "section-table";
        table.tabIndex = -1;
        table.id = sectId + "_attributes_table";
        
        var col_width = Math.floor(100 / nCols) + "%";
        var rowNum = 0;
        var atAttr = 0;        
        var advanceNextRow = true;
        var colNum = 0;
        var tr; 
        
        while(atAttr < nAttrs) {            
            if(advanceNextRow) {
                tr = table.insertRow(rowNum);
                tr.tabIndex = -1;
                rowNum++;
                colNum = 0;
                
                // reset the next row advancement
                advanceNextRow = false;
            }
            
            var attrKey = attrKeys[atAttr];
            var attrInfo = this.provider.getAttributeInfo(this.content, sectionKey, attrKey);
            var td = tr.insertCell(colNum);
            var colSpan = 1;
            
            td.tabIndex = -1;
            
            // apply the default "collection occupies a row"
            //
            if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_COLLECTION) {
                colSpan = nCols;
            }
            // if it has an override column span value, honor it. 
            if (attrInfo.colSpan) {
                colSpan = Math.min(nCols, attrInfo.colSpan);
            }
            
            // update the current column.
            colNum += colSpan;
                        
            // this attribute has a variable colspan, honor it.
            if(colSpan > 1) {
                td.colSpan = nCols;
            }
            else {
                // this is "normal" attribute. Make it a fixed
                // width to the number of column proportion.
                td.width = col_width;                   
            }
            
            // create a separator between collection attributes
            // but don't create one on the first item.
            if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_COLLECTION && attrInfo.separator) {
                tr = table.insertRow(rowNum++);
                td = tr.insertCell(0);
                var sep = document.createElement("div");
                sep.className = "collection-separator";
                td.appendChild(sep);
                td.colSpan = colSpan;       
            }
            
            var attrNode = this._createAttribute(td, sectionKey, attrKey);
            td.appendChild(attrNode);

            atAttr++;
            
            // we're already at the last column, advance to next row.
            if(colNum == nCols) {
                advanceNextRow = true;
                continue;
            }
            
            // check the next column attribute to see if it'll fit
            // in this row. If not, we'll have to wrap. But only do it
            // if we have not reach the end of attributes.
            // 
            if(atAttr < nAttrs) {
                var nextAttrId = attrKeys[atAttr];
                var nextAttrInfo = this.provider.getAttributeInfo(this.content, sectionKey, nextAttrId);
                var nextColSpan = 1;
                
                // apply the default "collection occupies a row"
                if(nextAttrInfo.type == SectionLayoutForm.ATTRIBUTE_COLLECTION) {
                    nextColSpan = nCols;
                }
                // if this attribute overrides the column span, honor it.
                if(nextAttrInfo.colSpan) {
                    nextColSpan = Math.min(nCols, nextAttrInfo.colSpan);
                }
                    
                // the next row don't fit, wrap it.
                if((nextColSpan + colNum) > nCols) {
                    advanceNextRow = true;
                }
            }
            else {
                // we've reached the last attribute, make it wrap now
                // to trigger filler columns check.
                advanceNextRow = true; 
            }
                    
            // create the filler cells in the row. This will force the previous cells
            // to occupy only the portion it's intended.
            if(advanceNextRow && (colNum < nCols)) {
                var nFillers = nCols - colNum;
                for(var i = 0; i < nFillers; i++) {
                    var td = tr.insertCell(i + colNum);
                    td.width = col_width;
                }
            }
        }

        return( table );
    },
    
    _createAttribute: function(rootNode, sectionKey, attributeKey)
    {
        var sectionInfo = this.provider.getSectionInfo(this.content, sectionKey);
        var attrInfo = this.provider.getAttributeInfo(this.content, sectionKey, attributeKey);
        var field = null;
        var fieldNode = null;
        var group = null;
        
        // if this attribute can be deleted, create root container.
        // We'll add a delete button after the field.
        //
        if(attrInfo.canDelete) {
            group = document.createElement("div");
            group.className = "field-root";
            group.tabIndex = -1;
            fieldNode = group;
        }
        
        if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_COLLECTION) {
            // if this field is not delete-able, we'll have to create
            // our own container for the collection.
            if(group == null) {
                group = document.createElement("div");
                group.className = "field-root";
                group.tabIndex = -1;
            }
            
            var groupWidget = null;
            var groupTitle = null;
            
            if(this.elementProvider != null) {
                groupWidget = this.elementProvider.constructCollectionAttributeHeader(sectionKey, attributeKey);
                
                if(groupWidget)
                    groupTitle = groupWidget.domNode;
            }
             
            if(attrInfo.title && groupTitle == null) {
                groupTitle = document.createElement("span");
                groupTitle.innerHTML = attrInfo.title;
                groupTitle.className = "collection-title";  
            }
            if(groupTitle) {
                group.appendChild(groupTitle);
            }
            var groupContainer = document.createElement("div");
            var level = (attrInfo.indentLevel && attrInfo.indentLevel < 5) ? attrInfo.indentLevel : "0";
            
            if(attrInfo.canDelete) {
                var buttonDiv = document.createElement("div");
                var deleteButton = this._createFieldDeleteButton(attributeKey, attrInfo);
                buttonDiv.appendChild(deleteButton);
                buttonDiv.className = "delete-button-area";
                
                if(groupWidget) {
                    dojo.addClass(groupContainer, "delete-attr-group-area");
                    groupWidget.contentNode.appendChild(buttonDiv);
                }
                else {
                    group.appendChild(buttonDiv);
                }
            }
            
            groupContainer.className = "collection-attribute-container";
            if(groupWidget) {
                // HACK - just have to know the collection widget has a content node.
                groupWidget.contentNode.appendChild(groupContainer);
            }
            else {
                group.appendChild(groupContainer);
                group.style.marginLeft = level + "em";
            }
            
            var collection = attrInfo.collection;
            var nCols = (attrInfo.nColumns) ? attrInfo.nColumns : sectionInfo.nColumns;
            
            // layout the collection of attributes.
            if(collection && dojo.isArray(collection)) {
                var node = this._createAttributesLayout(sectionKey, collection, collection.length, nCols);
                groupContainer.appendChild(node);               
            }
            fieldNode = group;
        }
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_TEXT_CHECKBOX) {
            field = new TextCheckboxField({fieldLabel: attrInfo.title, tooltipMessage:attrInfo.tooltipMessage});
            attrInfo.checkbox = field.getCheckboxReference();
        }
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_TEXT) {
            field = new TextField({fieldLabel: attrInfo.title});
            var value = attrInfo.initValue;
            if(value){
                field.setValue(value);
            }
        }
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_TEXT_NUMERIC) {
            field = new TextField({fieldLabel: attrInfo.title}); // numeric value is validated by backend javascript
        }        
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_TEXT_IP) {
            field = new IPAddressField({fieldLabel: attrInfo.title});
        }       
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_ENUM) {
            var needEmpty = (attrInfo.createFirstEmpty) ? true : false;
			var isDisabled = (attrInfo.disabled) ? true : false;
            field = new EnumField({fieldLabel: attrInfo.title, createEmpty: needEmpty, disabled:isDisabled});
            if(attrInfo.enumChoices) {
                field.setEnumChoices(attrInfo.enumChoices);
            }
        }
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_CUSTOM && this.elementProvider) {
            field = this.elementProvider.constructCustomField(attributeKey);
        }
        else if(attrInfo.type == SectionLayoutForm.ATTRIBUTE_TEXT_DATE) {
            field = new DateField({fieldLabel: attrInfo.title}); 
        }        
        else {
            // By default, if unknown type, or no custom element provider,
            // create a standard label field.
            field = new LabelField({fieldLabel: attrInfo.title});
        }
        
        // configure the field. Add the attributeKey to its userData member
        // for convenient access and hook into its value Changed callback.
        if(field) {
            // the field node is not assigned yet. We don't have an outer
            // container div. We can just return the field's main node.
            if (fieldNode == null) {
                fieldNode = field.domNode;
            }
            else {
                if(attrInfo.canDelete) {
                    fieldNode.appendChild(field.domNode);
                    var deleteButton = this._createFieldDeleteButton(attributeKey);
                    fieldNode.appendChild(deleteButton);
                }    
            }
            if(attrInfo.isRequired) {
                field.setIsRequired(true);
            }
            field.userData = attributeKey;
            field.onValueChanged = dojo.hitch(this, "_onFieldValueChanged");
        }
        
        // create the attribute data for the map
        var attrData = this._createLayoutData(rootNode, attributeKey, attrInfo, sectionKey, field);
        this.attributeMap[attrData.id] = attrData;
        
        return fieldNode;
    },    
    
    _createFieldDeleteButton: function(attributeKey, attrInfo)
    {
        var button = document.createElement("button");
        var img = document.createElement("img");
        var title = attrInfo.deleteTooltip ? attrInfo.deleteTooltip :  RtlmUiUtil.getMessage("labelDeleteButton");
        
        button.className = "delete-button-node";
        button.title = title;
        img.src = dojo.moduleUrl("com.ibm.rational.test.lm.web", "ui/internal/images/icons/dlcl16/delete.gif");
        button.appendChild(img);
        
        button.onmouseover = function() {
            this.image.src = dojo.moduleUrl("com.ibm.rational.test.lm.web", "ui/internal/images/icons/elcl16/delete.gif");
        }
        button.onmouseout = function() {
            this.image.src = dojo.moduleUrl("com.ibm.rational.test.lm.web", "ui/internal/images/icons/dlcl16/delete.gif");
        }
        
        // keep the attribute associated with this delete button
        // in the button.
        button.parentForm = this; 
        button.attrKey = attributeKey;
        button.image = img;
        button.onclick = function() {
            this.parentForm._onDeleteButtonClicked(this.attrKey);
        }
        return button;
    },
    
    _onDeleteButtonClicked: function(attributeKey)
    {
        // notify our listener about value changed in field.
        if(this.formListener) {
            this.formListener.attributeDeleteActionInvoked(attributeKey);
        }        
    },
    
    _onFieldValueChanged: function(fieldWidgetRef, oldValue, newValue)
    {
        // in loading value, ignore this and don't report to
        // the listener as value changed.
        if(this.inLoading)
            return;
            
        // notify our listener about value changed in field.
        if(this.formListener) {
            // we've stored the field's attributeKey into the widget's userData member
            // Just retrieve it and invoke the listener's message. 
            var attrKey = fieldWidgetRef.userData;
            var actions = this.formListener.attributeValueChanged(attrKey, oldValue, newValue);
            
            if(actions && dojo.isObject(actions)) {
                // The value change event could trigger a rebuild of the attribute/field
                // itself. Retrieve the layout record for the attribute again to get the
                // current field object
                var id = this._getKeyId(attrKey);
                var attrRec = this.attributeMap[id];
                
                if(attrRec) {
                    var field = attrRec[MapStructField];
                    if(field && actions["edited"] != "undefined")
                        field.setEdited(actions.edited);
                }
            }
        }
    },
    
    // Summary:
    //    For field validation that is performed by the application instead of the
    //    text field widget, it needs to set the text field status within the
    //    attribute value changed handler.
    setAttributeFieldStatus: function(attributeKey, actions)
    {
        if(actions && dojo.isObject(actions)) {
            var id = this._getKeyId(attributeKey);
            var attrRec = this.attributeMap[id];
        
            if(attrRec) {
                var field = attrRec[MapStructField];
                   if (field) {
                    field.setHasError(actions["hasError"] != undefined);
                    
                    if (actions["hasError"] != undefined) {
                        field.setErrorMessage(actions["errorMessage"]);
                    } else if (actions["edited"] != undefined) {
                        field.setEdited(actions.edited);
                    }
                }
            }
        }
    },
    
    _getKeyId: function(key)
    {
        var id;
        
        if(dojo.isString(key)) {
            id = key;
        }
        else if(dojo.isObject(key) && key.id) {
            id = key.id;
        }
        else {
            throw new Error("Layout key needs to be a string, or a struct with an \'id\' attribute.");
        }
        return id;      
    },

    // Summary:
    //    Creates a layout record to associate a key, either a sectionKey, or an
    //    attributeKey to its UI object. For an attribute, it's the field.
    // 
    _createLayoutData: function(rootNode, key, info, secondKey, field)
    {
        var data = {};
        var id = this._getKeyId(key);
        
        data[MapStructId] = id;
        data[MapStructInfo] = info; //this._copyStruct(info);
        data[MapStructRootNode] = rootNode;
        data[MapStructObjectKey] = key; //this._copyStruct(key);
        
        if(secondKey) {
            data[MapStructObjectSecondKey] = secondKey;//this._copyStruct(secondKey);
        }
        if(field)
            data[MapStructField] = field;
        return data;
    },
    
    _copyStruct: function(obj)
    {
        if(!dojo.isObject(obj))
            return obj;
            
        var dup = {};
        
        for(var attr in obj) {
            if(dojo.isObject(obj[attr])) {
                dup[attr] = this._copyStruct(obj[attr]);
            }
            else {
                dup[attr] = obj[attr];                
            }
        }
        return dup;
    }
    
});

})();
