/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.rational.test.lm.spi.util.topology;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.sql.Timestamp;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.XMLResource;

import com.ibm.ccl.soa.deploy.core.CoreFactory;
import com.ibm.ccl.soa.deploy.core.DeployCoreRoot;
import com.ibm.ccl.soa.deploy.core.ExtendedAttribute;
import com.ibm.ccl.soa.deploy.core.HostingLink;
import com.ibm.ccl.soa.deploy.core.InstallState;
import com.ibm.ccl.soa.deploy.core.Topology;
import com.ibm.ccl.soa.deploy.core.Unit;
import com.ibm.ccl.soa.deploy.genericsoftware.GenericsoftwareFactory;
import com.ibm.ccl.soa.deploy.genericsoftware.SoftwareInstall;
import com.ibm.ccl.soa.deploy.genericsoftware.SoftwareInstallUnit;
import com.ibm.ccl.soa.deploy.net.IPv6AddressNotationType;
import com.ibm.ccl.soa.deploy.net.IpInterface;
import com.ibm.ccl.soa.deploy.net.IpInterfaceUnit;
import com.ibm.ccl.soa.deploy.net.NetFactory;
import com.ibm.ccl.soa.deploy.operation.Operation;
import com.ibm.ccl.soa.deploy.operation.OperationFactory;
import com.ibm.ccl.soa.deploy.operation.OperationUnit;
import com.ibm.ccl.soa.deploy.operation.RunState;
import com.ibm.ccl.soa.deploy.os.KernelWidthType;
import com.ibm.ccl.soa.deploy.os.OperatingSystem;
import com.ibm.ccl.soa.deploy.os.OperatingSystemUnit;
import com.ibm.ccl.soa.deploy.os.OsFactory;
import com.ibm.ccl.soa.deploy.server.ArchitectureWidthType;
import com.ibm.ccl.soa.deploy.server.Server;
import com.ibm.ccl.soa.deploy.server.ServerFactory;
import com.ibm.ccl.soa.deploy.server.ServerUnit;
import com.ibm.ccl.soa.deploy.virtualization.VirtualImage;
import com.ibm.ccl.soa.deploy.virtualization.VirtualImageCollection;
import com.ibm.ccl.soa.deploy.virtualization.VirtualImageCollectionUnit;
import com.ibm.ccl.soa.deploy.virtualization.VirtualImageUnit;
import com.ibm.ccl.soa.deploy.virtualization.VirtualizationFactory;

/**
 * 
 */
public class TopologyCreator {
    private static final String UTF8Encoding = "UTF-8";
    
    // Core Units don't have the RunState, externalId property, imageType, and 
    // imageId so we will add four extended Attributes.
    public final static String RUN_STATE = "CurrentRunState";//$NON-NLS-1$
    public final static String EXTERNAL_ID = "VirtualMachineCollectionUnit_ExternalId";//$NON-NLS-1$
    public final static String IMAGE_ID = "ImageId"; //$NON-NLS-N$
    public final static String IMAGE_TYPE = TopologyConstant.virtualImageImageTypeAttribute;

    /**
     * Create Server Unit
     * 
     * @param unitName
     * @param serverName
     * @param CPUArch
     * @param CPUCount
     * @param MemorySize
     * @return
     */
    @SuppressWarnings("unchecked")
    public static ServerUnit createServerUnit(String serverName, String systemID) {

        ServerUnit serverUnit = ServerFactory.eINSTANCE.createServerUnit();
        serverUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);

        Server server = ServerFactory.eINSTANCE.createServer();
        server.setName(serverName);
        if ((systemID != null) && (!"".equals(systemID))) {//$NON-NLS-1$
            server.setSystemId(systemID);
        }
        serverUnit.getHostingOrAnyCapabilities().add(server);
        return serverUnit;
    }
    
    /**
     * Create a TDM ServerUnit containing provided information
     * @param name
     * @param isVirtual
     * @param description
     * @param manufacturer
     * @param memory
     * @param model
     * @param primaryMacAddress
     * @param serialNumber
     * @param systemBoardUUID
     * @param systemId
     * @return
     */
    @SuppressWarnings("unchecked")
    public static ServerUnit createServerUnit(
            String name,
            boolean isVirtual,
            String description,
            String manufacturer,
            BigInteger memory,            
            String model,            
            String primaryMacAddress,
            String serialNumber,
            String systemBoardUUID,
            String systemId) {
        Server server = ServerFactory.eINSTANCE.createServer();        
        server.setVirtual(isVirtual);        
        server.setDescription(description);
        server.setManufacturer(manufacturer);
        server.setMemorySize(memory);
        server.setModel(model);
        server.setName(name);
        server.setPrimaryMACAddress(primaryMacAddress);
        server.setSerialNumber(serialNumber);
        server.setSystemBoardUUID(systemBoardUUID);
        server.setSystemId(systemId);                

        ServerUnit serverUnit = ServerFactory.eINSTANCE.createServerUnit();
        serverUnit.setName(name);
        serverUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);        
        serverUnit.getHostingOrAnyCapabilities().add(server);
        return serverUnit;        
    }
    
    /**
     * Add CPU information to an existing TDM ServerUnit
     * @param serverUnit
     * @param architecture
     * @param architectureWidth
     * @param coresInstalled
     * @param count
     * @param diesInstalled
     * @param manufacturer
     * @param speed
     * @param type
     * @return
     */
    @SuppressWarnings("unchecked")
    public static ServerUnit addCPUInfoToServer(
            ServerUnit serverUnit,
            String architecture,
            String architectureWidth,
            Integer coresInstalled,
            Integer count,
            Integer diesInstalled,
            String manufacturer,
            Integer speed,
            String type) {
        List<Object> hostedList = serverUnit.getHostingOrAnyCapabilities();
        for (Object hosted : hostedList) {
            if (hosted instanceof Server) {
                ((Server)hosted).setCpuArchitecture(architecture);
                if (architectureWidth != null && architectureWidth.length() > 0) {
                    if (architectureWidth.contains("64")) { //$NON-NLS-N$
                        ((Server)hosted).setCpuArchitectureWidth(ArchitectureWidthType._64_BIT_LITERAL);
                    } else {
                        ((Server)hosted).setCpuArchitectureWidth(ArchitectureWidthType._32_BIT_LITERAL);
                    }
                }
                if (coresInstalled != null) {
                    ((Server)hosted).setCpuCoresInstalled(coresInstalled);
                }
                if (count != null) {
                    ((Server)hosted).setCpuCount(BigInteger.valueOf(count));
                }
                if (diesInstalled != null) {                
                    ((Server)hosted).setCpuDiesInstalled(diesInstalled);
                }
                ((Server)hosted).setCpuManufacturer(manufacturer);
                if (speed != null) {
                    ((Server)hosted).setCpuSpeed(speed);
                }
                ((Server)hosted).setCpuType(type);
            }
        }
        return serverUnit;
    }
    
    /**
     * Create a TDM OperatingSystemUnit containing the provided information
     * @param fqdn
     * @param hostname
     * @param kernelArch
     * @param kernelVersion
     * @param kernelWidth
     * @param locale
     * @param manufacturer
     * @param distribution
     * @param servicepack
     * @param type
     * @param version
     * @param timezone
     * @return
     */
    @SuppressWarnings("unchecked")
    public static OperatingSystemUnit createOsUnit(
            String fqdn,
            String hostname,
            String kernelArch,
            String kernelVersion,
            String kernelWidth,
            String locale,
            String manufacturer,
            String distribution,
            String servicepack,
            String type,
            String version,
            String timezone) {
        OperatingSystem os = OsFactory.eINSTANCE.createOperatingSystem();        
        os.setFqdn(fqdn);
        os.setHostname(hostname);
        os.setKernelArchitecture(kernelArch);
        os.setKernelVersion(kernelVersion);
        if (kernelWidth != null && kernelWidth.length() > 0) {
            if (kernelWidth.contains("64")) {
                os.setKernelWidth(KernelWidthType._64_BIT_LITERAL);
            } else {
                os.setKernelWidth(KernelWidthType._32_BIT_LITERAL);
            }
        }
        os.setLocale(locale);
        os.setManufacturer(manufacturer);
        os.setOsDistribution(distribution);
        os.setOsServicepack(servicepack);
        os.setOsType(type);
        os.setOsVersion(version);
        os.setTimeZone(timezone);
        
        OperatingSystemUnit osUnit = OsFactory.eINSTANCE.createOperatingSystemUnit();
        osUnit.setName(type);
        osUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);
        osUnit.getHostingOrAnyCapabilities().add(os);
        return osUnit;
    }           
    
    /**
     * Create a TDM IpInterfaceUnit containing the provided information
     * @param ipV4Address
     * @param ipV6Address
     * @param ipV6NotationType
     * @param l2InterfaceName
     * @param netmask
     * @param subnetAddress
     * @return
     */
    @SuppressWarnings("unchecked")
    public static IpInterfaceUnit createIpInterfaceUnit(
            String ipV4Address,
            String ipV6Address,
            String ipV6NotationType,
            String l2InterfaceName,
            String netmask,
            String subnetAddress) {
        IpInterface ipInterface = NetFactory.eINSTANCE.createIpInterface();
        ipInterface.setIpV4Address(ipV4Address);
        ipInterface.setIpV6Address(ipV6Address);
        if (ipV6NotationType != null && ipV6NotationType.equalsIgnoreCase("Dotted")) {//$NON-NLS-N$
            ipInterface.setIpv6NotationType(IPv6AddressNotationType.IPV6_MIXED_NOTATION_LITERAL);
        } else if (ipV6NotationType != null && ipV6NotationType.equalsIgnoreCase("Hex")) {//$NON-NLS-N$
            ipInterface.setIpv6NotationType(IPv6AddressNotationType.IPV6_COLON_NOTATION_LITERAL);
        }        
        ipInterface.setL2InterfaceName(l2InterfaceName);
        ipInterface.setNetmask(netmask);
        ipInterface.setSubnetAddress(subnetAddress);
        
        IpInterfaceUnit ipInterfaceUnit = NetFactory.eINSTANCE.createIpInterfaceUnit();
        if (ipV4Address != null && ipV4Address.length() > 0) {
            ipInterfaceUnit.setName(ipV4Address);
        } else if (ipV6Address != null && ipV6Address.length() > 0) {
            ipInterfaceUnit.setName(ipV6Address);
        } else {
            ipInterfaceUnit.setName("IpInterface" + UUID.randomUUID().toString()); //$NON-NLS-N$
        }
        ipInterfaceUnit.getHostingOrAnyCapabilities().add(ipInterface);
        return ipInterfaceUnit;
    }
    
    /**
     * Create a single server topology and link descendant resources (ip, os, etc)
     * to the ServerUnit
     * @param serverUnit
     * @param osUnits
     * @param ipUnits
     * @return
     */
    @SuppressWarnings("unchecked")
    public static Topology createServerTopology(
            ServerUnit serverUnit, 
            List<OperatingSystemUnit> osUnits,
            List<IpInterfaceUnit> ipUnits) {
        Topology topology = CoreFactory.eINSTANCE.createTopology();
        topology.setDisplayName("Server Topology");
        topology.getUnits().add(serverUnit);
        for (OperatingSystemUnit osUnit : osUnits) {
            topology.getUnits().add(osUnit);
            createHostLink(serverUnit, osUnit);
            for (IpInterfaceUnit ipUnit : ipUnits) {
                if (!topology.getUnits().contains(ipUnit)) {
                    topology.getUnits().add(ipUnit);
                }
                createHostLink(osUnit, ipUnit);
            }
        }
        return topology;
    }
    
    /**
     * Serialize a TDM Topology to its XML String representation
     * @param topology
     * @return
     */
    public static String topologyToString(Topology topology) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        saveTopologyToStream(topology, baos);
        String xml = null;
        try {
            xml = baos.toString(UTF8Encoding);
            return xml;
        } catch (UnsupportedEncodingException e) {
            return null;
        }        
    }

    /**
     * Create Job Definition Unit
     * 
     * @param unitName
     * @param externalId
     * @param jobName
     * @return
     */
    public static OperationUnit createJobDefUnit(String unitName,
            String externalId, String jobName) {
        OperationUnit jobDefUnit = OperationFactory.eINSTANCE
                .createOperationUnit("Jobs", null,//$NON-NLS-1$
                        new String[] {}, null);//$NON-NLS-1$
        jobDefUnit.setName(unitName);
        assert jobDefUnit != null;

        Operation jobDefOp = jobDefUnit.getOperation();
        // jobDef.setAction(InstallerAction.INSTALL_LITERAL);
        jobDefOp.setOperationName(jobName);
        jobDefOp.setOperationId(String.valueOf(externalId));

        return jobDefUnit;
    }

    /**
     * Create a unit representing a piece of software
     * @deprecated please use {@link #createSoftwareInstallUnit(String, String, String, String, String, String, String, String, String, String, String, String, String)}

     * @param buildVersion SW build
     * @param description textual description of the software
     * @param execName name of the executable
     * @param installDate Timestamp value converted to number of milliseconds 
     * 	since January 1, 1970, 00:00:00
     * @param installedLocation path to install directory
     * @param productId 
     * @param productName Name of the product, not constrained to catalog values like swName
     * @param publisher publisher of the SW
     * @param displayName human readable name 
     * @param password optional password info for software. NOT encrypted.
     * @param user optional username info for software
     * @param dirPath optional working directory for software. Typically used for agent software.
     * @return
     */
    @Deprecated
    public static SoftwareInstallUnit createSoftwareInstallUnit(
            String buildVersion, String description, String execName,
            String installDate, String installedLocation, String productId,
            String productName, String publisher, String displayName,
            String password, String user, String dirPath) {
    	return createSoftwareInstallUnit(null, buildVersion, description, execName,
    			installDate, installedLocation, productId, productName, publisher,
    			displayName, password, user, dirPath);
    }
    
    /**
     * Create a unit representing a piece of software
     * @param swName the type name of the Software. Only used if mapped to a value  
     * 	in the LM software catalog.
     * @param buildVersion SW build
     * @param description textual description of the software
     * @param execName name of the executable
     * @param installDate Timestamp value converted to number of milliseconds 
     * 	since January 1, 1970, 00:00:00
     * @param installedLocation path to install directory
     * @param productId 
     * @param productName Name of the product, not constrained to catalog values like swName
     * @param publisher publisher of the SW
     * @param displayName human readable name 
     * @param password optional password info for software. NOT encrypted.
     * @param user optional username info for software
     * @param dirPath optional working directory for software. Typically used for agent software.
     * @return
     */
    @SuppressWarnings("unchecked")
    public static SoftwareInstallUnit createSoftwareInstallUnit(
            String swName, String buildVersion, String description, String execName,
            String installDate, String installedLocation, String productId,
            String productName, String publisher, String displayName,
            String password, String user, String dirPath) {
        SoftwareInstallUnit softwareInstallUnit = GenericsoftwareFactory.eINSTANCE
                .createSoftwareInstallUnit();//$NON-NLS-1$
        assert softwareInstallUnit != null;
        SoftwareInstall sw_Install = GenericsoftwareFactory.eINSTANCE
                .createSoftwareInstall();
        sw_Install.setName(swName);
        sw_Install.setBuildVersion(buildVersion);
        sw_Install.setDescription(description);
        sw_Install.setExecutableName(execName);
        sw_Install.setInstallDate(installDate);
        sw_Install.setInstalledLocation(installedLocation);
        sw_Install.setProductId(productId);
        sw_Install.setProductName(productName);
        sw_Install.setDisplayName(displayName);
        sw_Install.setPublisher(publisher);
        if (password != null) {
            ExtendedAttribute password_EA = CoreFactory.eINSTANCE
                    .createExtendedAttribute();
            password_EA.setName(TopologyConstant.softwarePasswordAttribute);//$NON-NLS-1$
            password_EA.setValue(password);
            sw_Install.getExtendedAttributes().add(password_EA);
        }
        if (user != null) {
            ExtendedAttribute user_EA = CoreFactory.eINSTANCE
                    .createExtendedAttribute();
            user_EA.setName(TopologyConstant.softwareUserNameAttribute);//$NON-NLS-1$
            user_EA.setValue(user);
            sw_Install.getExtendedAttributes().add(user_EA);
        }
        if (dirPath != null) {
            ExtendedAttribute dirPath_EA = CoreFactory.eINSTANCE
                    .createExtendedAttribute();
            dirPath_EA.setName(TopologyConstant.softwareTempPathAttribute);//$NON-NLS-1$
            dirPath_EA.setValue(dirPath);
            sw_Install.getExtendedAttributes().add(dirPath_EA);
        }
        softwareInstallUnit.getHostingOrAnyCapabilities().add(sw_Install);
        return softwareInstallUnit;
    }

    
    /**
	 * Create software install unit with extended attributes in each of which
	 * the name is the key of attributesMap and the value is the value of
	 * attributesMap
	 * 
	 * @param attributesMap
	 * @return
	 */
    @SuppressWarnings("unchecked")
	public static SoftwareInstallUnit createSWInstallUnit(
			Map<String, String> attributesMap) {
		SoftwareInstallUnit swInstallUnit = GenericsoftwareFactory.eINSTANCE
				.createSoftwareInstallUnit();
		SoftwareInstall swInstall = GenericsoftwareFactory.eINSTANCE
				.createSoftwareInstall();
		for (Map.Entry<String, String> entry : attributesMap.entrySet()) {
			ExtendedAttribute attribute = CoreFactory.eINSTANCE
					.createExtendedAttribute();
			attribute.setName(entry.getKey());
			attribute.setValue(entry.getValue());
			swInstall.getExtendedAttributes().add(attribute);
		}
		swInstallUnit.getHostingOrAnyCapabilities().add(swInstall);
		return swInstallUnit;
	}
    
    /**
     * Create Windows OS Unit
     * 
     * @param osName
     * @param osType
     * @param hostName
     * @return
     */
    @SuppressWarnings("unchecked")
    public static OperatingSystemUnit createOSUnit(String osName, String osType, String hostName) {

        OperatingSystem os = OsFactory.eINSTANCE.createOperatingSystem();
        os.setName(osName);
        os.setOsType(osType);
        os.setHostname(hostName);

        OperatingSystemUnit osUnit = OsFactory.eINSTANCE
                .createOperatingSystemUnit();
        osUnit.setName(osName);
        osUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);
        osUnit.getHostingOrAnyCapabilities().add(os);

        return osUnit;
    }
    
    /**
     * Create IP interface Unit
     * 
     * @param ipAddress
     *            IP address value (V4 or V6)
     * @param isV6
     *            true - V6 IP; false - V4 IP
     * @return Created IpInterfaceUnit
     */
    @SuppressWarnings("unchecked")
    public static IpInterfaceUnit createIpInterfaceUnit(String ipAddress, boolean isV6) {
        IpInterfaceUnit ipInterfaceUnit = NetFactory.eINSTANCE
                .createIpInterfaceUnit();
        ipInterfaceUnit.setName("IP"); //$NON-NLS-N$
        IpInterface ipInterface = NetFactory.eINSTANCE.createIpInterface();
        if (isV6) {
            ipInterface.setIpV6Address(ipAddress);
            // Set notation type to colon
            ipInterface
                    .setIpv6NotationType(IPv6AddressNotationType.IPV6_COLON_NOTATION_LITERAL);
        } else {
            ipInterface.setIpV4Address(ipAddress);
        }
        ipInterfaceUnit.getHostingOrAnyCapabilities().add(ipInterface);
        return ipInterfaceUnit;
    }

    /**
     * Create hostLink between the host and hostee
     * 
     * @param host
     * @param hostee
     */
    @SuppressWarnings("unchecked")
    public static void createHostLink(Unit host, Unit hostee) {
        HostingLink hostingLink = CoreFactory.eINSTANCE.createHostingLink();
        hostingLink.setHosted(hostee);
        hostingLink.setName(host.getName() + "_hosting_" + hostee.getName());
        host.getHostingLinks().add(hostingLink);
    }

    /**
     * 
     * @param m_topology
     * @param os
     */
    @SuppressWarnings("unchecked")
    public static void saveTopologyToStream(Topology m_topology, OutputStream os) {
        Resource res = m_topology.eResource();

        if (res == null) {
            // Topology isn't yet in a resource; make one for it. (If Topology
            // was in a resource it wouldn't be safe to add it to a new one
            // because that would
            // remove it from its current Resource).
            ResourceSet rs = new ResourceSetImpl();
            res = rs.createResource(URI.createURI("resource.extension"));
            DeployCoreRoot root = CoreFactory.eINSTANCE.createDeployCoreRoot();
            root.setTopology(m_topology);
            res.getContents().add(root);
        }

        try {
            HashMap options = new HashMap();
            options.put(XMLResource.OPTION_EXTENDED_META_DATA, Boolean.TRUE);
            res.save(os, options);
        } catch (IOException ex) {
            // needs logging...
            ex.printStackTrace();
        } catch (IllegalArgumentException ex) {
            ex.printStackTrace();
        }
    }
    
    /**
     * Create virtual Collections topology from virtual image colletion unit
     * list.
     * 
     * @param virtualImageCollectionUnits
     * @return virtual machines topology string
     */
    public static String createVirtualImageCollectionTopology(
            List<VirtualImageCollectionUnit> virtualImageCollectionUnits) {

        Topology top = CoreFactory.eINSTANCE.createTopology();
        top.setDisplayName("Virtual Iamge Collection Topology");//$NON-NLS-1$
        for (VirtualImageCollectionUnit virtualImageCollectionUnit : virtualImageCollectionUnits) {
            addHostedUnits(top, virtualImageCollectionUnit);
        }
        return topologyToString(top);
    }
    
    /**
     * Create virtual images topology from virtual image unit
     * 
     * @param virtualImageUnits
     *            A list of virtual image unit
     * @return virtual images topology string
     */
    public static String createVirtualImagesTopology(
            List<VirtualImageUnit> virtualImageUnits) {

        Topology top = CoreFactory.eINSTANCE.createTopology();
        top.setDisplayName("Virtual Image Topology");//$NON-NLS-1$
        for (VirtualImageUnit virtualImageUnit : virtualImageUnits) {
            addHostedUnits(top, virtualImageUnit);
        }
        return topologyToString(top);
    }
    
    /**
     * Create virtual machines topology from server unit list
     * 
     * @param serverUnits
     *            A list of server unit
     * @return virtual machines topology string
     */
    public static String createVirtualMachinesTopology(
            List<ServerUnit> serverUnits) {

        Topology top = CoreFactory.eINSTANCE.createTopology();
        top.setDisplayName("Virtual Machines Topology");//$NON-NLS-1$

        for (ServerUnit serverUnit : serverUnits) {
            addHostedUnits(top, serverUnit);
        }
        return topologyToString(top);
    }
    
    /*
     * add unit and recursively add all hosted units to the topology 
     */
    @SuppressWarnings("unchecked")
    private static void addHostedUnits(Topology topology, Unit unit) {
        topology.getUnits().add(unit);
        List hostingLinks = unit.getHostingLinks();
        Iterator<HostingLink> iterator = hostingLinks.iterator();
        while (iterator.hasNext()) {
            Unit hosted = iterator.next().getHosted();
            if (hosted != null) {
                addHostedUnits(topology, hosted);
            }
        }
    }
    
    /**
     * Create a collection of virtual images
     * @deprecated Please use {@link #createVirtualImageCollectionUnit(String, String, String)}
     * 
     * @param name name of the collection
     * @param virtualCollectionId id of the collection
     * @return
     */
    @Deprecated
    public static VirtualImageCollectionUnit createVirtualImageCollectionUnit(
            String name, String virtualCollectionId) {
    	return createVirtualImageCollectionUnit(name, virtualCollectionId, "");
    }
    
    /**
     * Create a collection of virtual images
     * 
     * @param name name of the collection
     * @param virtualCollectionId id of the collection
     * @param description description of collection
     * @return
     */
    @SuppressWarnings("unchecked")
    public static VirtualImageCollectionUnit createVirtualImageCollectionUnit(
            String name, String virtualCollectionId, String description) {
        VirtualImageCollectionUnit viCollectionUnit = VirtualizationFactory.eINSTANCE
                .createVirtualImageCollectionUnit();
        viCollectionUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);
        viCollectionUnit.setName(name);
        viCollectionUnit.setDescription(description);

        VirtualImageCollection virtualImageCollection = VirtualizationFactory.eINSTANCE
                .createVirtualImageCollection();
        virtualImageCollection.setCollectionId(virtualCollectionId);
        virtualImageCollection.setName(name);
        viCollectionUnit.getHostingOrAnyCapabilities().add(
                virtualImageCollection);
        return viCollectionUnit;
    }
    
    /**
     * Change the RunState Extended Attribute in the virtual machine unit.
     * 
     * @param unit vm unit            
     * @param runState current vm status
     */
    @SuppressWarnings("unchecked")
    public static void changeVMRunState(Unit unit, RunState runState) {
        List<ExtendedAttribute> attributes = unit.getExtendedAttributes();        
        for (int i=0; i<attributes.size(); i++) {
            ExtendedAttribute attribute = attributes.get(i);
            if (RUN_STATE.equals(attribute.getName())) {
                attributes.remove(attribute);
            }
        }        
        addRunState(unit, runState);
    }
    
    /**
     * Add RunState into unit.
     * 
     * @param unit
     * @param runState
     */
    @SuppressWarnings("unchecked")
    private static void addRunState(Unit unit, RunState runState) {
        ExtendedAttribute runState_EA = CoreFactory.eINSTANCE
                .createExtendedAttribute();
        runState_EA.setName(RUN_STATE);
        runState_EA.setValue(runState.getLiteral());
        unit.getExtendedAttributes().add(runState_EA);
    }
    
    /**
     * Create virtual Collections topology from virtual machine collection unit
     * list
     * Each machine collection unit need to link with the deployed image collection unit.
     * Each machine collection unit need to link with its machine units.
     * Each machine unit need to link with its deployed image unit.
     * @param virtualMachineCollectionUnits
     * @return virtual machines topology string
     */
    public static String createVirtualMachineCollectionTopology(
            List<Unit> virtualMachineCollectionUnits) {

        Topology top = CoreFactory.eINSTANCE.createTopology();
        top.setDisplayName("Virtual Machine Collection Topology");//$NON-NLS-1$
        for (Unit virtualMachineCollectionUnit : virtualMachineCollectionUnits) {
            addHostedUnits(top, virtualMachineCollectionUnit);
        }
        return topologyToString(top);
    }
    
    /**
     * Create a virtual image unit
     * @deprecated Please use {@link #createVirtualImageUnit(String, String, String, String, String, String, Timestamp, Timestamp, OperatingSystemUnit)}
     * 
     * @param name name of the image
     * @param identifier id of the image
     * @param toolName tool that created the image
     * @param buildVersion
     * @param dateCreated
     * @param dateModified
     * @param osUnit optional Operating System information
     * @return
     */
    @Deprecated
    public static VirtualImageUnit createVirtualImageUnit(
            String name, 
            String identifier, 
            String toolName, 
            String buildVersion,
            Timestamp dateCreated,
            Timestamp dateModified,
            OperatingSystemUnit osUnit) {
    	return createVirtualImageUnit(name, identifier, "", "", toolName, buildVersion, dateCreated, dateModified, osUnit);
    }
    
    /**
     * Create a virtual image unit
     * 
     * @param name name of the image
     * @param identifier id of the image
     * @param description textual description of image
     * @param imageType enumerated image type
     * @param toolName tool that created the image
     * @param buildVersion
     * @param dateCreated
     * @param dateModified
     * @param osUnit optional Operating System information
     * @return
     */
    @SuppressWarnings("unchecked")
    public static VirtualImageUnit createVirtualImageUnit(
            String name, 
            String identifier,
            String description,
            String imageType,
            String toolName, 
            String buildVersion,
            Timestamp dateCreated,
            Timestamp dateModified,
            OperatingSystemUnit osUnit) {

        VirtualImageUnit imageUnit = VirtualizationFactory.eINSTANCE
                .createVirtualImageUnit();
        imageUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);
        imageUnit.setName(name);

        VirtualImage image = VirtualizationFactory.eINSTANCE
                .createVirtualImage();
        image.setName(name);
        image.setDescription(description);
        image.setBuildVersion(buildVersion);
        image.setDateCreated(Long.toString(dateCreated.getTime()));
        image.setDateLastModified(Long.toString(dateModified.getTime()));
        image.setCreatedWithVMTool(toolName);
        
        ExtendedAttribute imageId_EA = CoreFactory.eINSTANCE
                .createExtendedAttribute();
        imageId_EA.setName(IMAGE_ID);
        imageId_EA.setValue(identifier);
        image.getExtendedAttributes().add(imageId_EA);
        
        ExtendedAttribute imageType_EA = CoreFactory.eINSTANCE
                .createExtendedAttribute();
        imageType_EA.setName(IMAGE_TYPE);
        imageType_EA.setValue(imageType);
        image.getExtendedAttributes().add(imageType_EA);

        imageUnit.getHostingOrAnyCapabilities().add(image);        
        if (osUnit != null) {
            TopologyCreator.createHostLink(imageUnit, osUnit);
        }

        return imageUnit;

    }

    /**
     * Create a virtual machine unit with descriptive information.
     * @deprecated Please use {@link #createVirtualMachineUnitForVMId(String, String, String, RunState, OperatingSystemUnit, IpInterfaceUnit)}
     *
     * @param virtualMachineName name of the VM
     * @param virtualMachineId id of the VM
     * @param runState the status of the VM (i.e. "running")
     * @param osUnit the OS information of the VM
     * @param ipInterfaceUnit the ip information of the VM
     * @return
     */
    @Deprecated
    public static ServerUnit createVirtualMachineUnitForVMId(
            String virtualMachineName, 
            String virtualMachineId, 
            RunState runState, 
            OperatingSystemUnit osUnit, 
            IpInterfaceUnit ipInterfaceUnit) {
    	return createVirtualMachineUnitForVMId(virtualMachineName, virtualMachineId, "", runState, osUnit, ipInterfaceUnit);
    }

    /**
     * Create a virtual machine unit with descriptive information.
     *
     * @param virtualMachineName name of the VM
     * @param virtualMachineId id of the VM
     * @param virtualMachineDescription textual description of the VM
     * @param runState the status of the VM (i.e. "running")
     * @param osUnit the OS information of the VM
     * @param ipInterfaceUnit the ip information of the VM
     * @return
     */
    @SuppressWarnings("unchecked")
    public static ServerUnit createVirtualMachineUnitForVMId(
            String virtualMachineName, 
            String virtualMachineId,
            String virtualMachineDescription,
            RunState runState, 
            OperatingSystemUnit osUnit, 
            IpInterfaceUnit ipInterfaceUnit) {
        ServerUnit serverUnit = ServerFactory.eINSTANCE.createServerUnit();
        serverUnit.setInitInstallState(InstallState.UNKNOWN_LITERAL);
        serverUnit.setName(virtualMachineName);
        TopologyCreator.changeVMRunState(serverUnit, runState);

        Server virtualMachine = ServerFactory.eINSTANCE.createServer();
        virtualMachine.setVirtual(true);
        virtualMachine.setSystemId(virtualMachineId);
        virtualMachine.setName(virtualMachineName);
        virtualMachine.setDescription(virtualMachineDescription);
        serverUnit.getHostingOrAnyCapabilities().add(virtualMachine);
        if (ipInterfaceUnit != null && osUnit != null) {
            TopologyCreator.createHostLink(osUnit, ipInterfaceUnit);
        }
        if (osUnit != null) {
            TopologyCreator.createHostLink(serverUnit, osUnit);
        }
        return serverUnit;
    }    
    
    /**
     * 
     * @param unit
     * @param externalId
     */
    @SuppressWarnings("unchecked")
    private static void addExternalId(Unit unit, String externalId) {
        ExtendedAttribute externalId_EA = CoreFactory.eINSTANCE
                .createExtendedAttribute();
        externalId_EA.setName(EXTERNAL_ID);//$NON-NLS-1$
        externalId_EA.setValue(externalId);
        unit.getExtendedAttributes().add(externalId_EA);
    }

    /**
     * Create a Unit used to store a collection of Virtual Machines.
     * @deprecated Please use {@link #createVirtualMachineCollectionUnit(String, String, String, RunState)}
     * 
     * @param virtualMachineCollectionName
     *            VM Collection name
     * @param virtualMachineCollectionId
     *            Machine Collection Id
     * @param runstate
     *            state of the virtual machine
     * 
     * @return
     */
    @Deprecated
    public static Unit createVirtualMachineCollectionUnit(
            String virtualMachineCollectionName, String virtualMachineCollectionId, RunState runstate) {
    	return createVirtualMachineCollectionUnit(virtualMachineCollectionName, virtualMachineCollectionId, "", runstate);
    }
    
    /**
     * Create a Unit used to store a collection of Virtual Machines.
     * 
     * @param virtualMachineCollectionName
     *            VM Collection name
     * @param virtualMachineCollectionId
     *            Machine Collection Id
     * @param vmcDescription 
     *            textual description of the virtual machine collection           
     * @param runstate
     *            state of the virtual machine
     * 
     * @return
     */
    public static Unit createVirtualMachineCollectionUnit(
            String virtualMachineCollectionName, String virtualMachineCollectionId, 
            String vmcDescription, RunState runstate) {
        Unit vmCollectionUnit = CoreFactory.eINSTANCE.createUnit();
        vmCollectionUnit.setDisplayName(virtualMachineCollectionName);
        vmCollectionUnit.setName(virtualMachineCollectionName);
        vmCollectionUnit.setDescription(vmcDescription);
        addExternalId(vmCollectionUnit, virtualMachineCollectionId);
        TopologyCreator.changeVMRunState(vmCollectionUnit, runstate);
        return vmCollectionUnit;
    }
}
